import math
import numpy as np
import os
import ctcsound
import tkinter as tk
from tkinter import filedialog
from sklearn.decomposition import NMF
import matplotlib.pyplot as plt
import copy
from random import randrange
from msvcrt import getch,kbhit
from itertools import combinations

#
# C-Sound parameters
#
SAMPLE_RATE = 44100
FFT_SIZE = 2048
SAMPLES_PER_K = FFT_SIZE

#
# MODEL parameters
#
COMBINED_COLUMNS = 10
COMPONENTS = 60
INSTRUMENTS = 4

#
# VERIFICATION-TEST parameters
#
TEST_TOTAL_TIME = 15            # seconds per single test
TEST_SAMPLE_DENSITY = 1.75       # Samples per second per instrument

#
# Misc. Constants
#
TABLE = 1

c = ctcsound.Csound()

instrumentsH = [[([1.0]*FFT_SIZE*COMBINED_COLUMNS) for i in range(COMPONENTS)] for j in range(INSTRUMENTS)]
initialized = [False]*INSTRUMENTS
componentMetrics = [([100.0]*INSTRUMENTS) for i in range(COMPONENTS*INSTRUMENTS)]

model = None

def prepare_model():
    global model
    comp = []
    for instrument in instrumentsH:
        for component in instrument:
            comp.append(component)
            
    if model == None:
        print("Making model, this unfortunately takes a little bit of time")
        model = NMF(n_components=COMPONENTS*INSTRUMENTS, init='custom')
        model.fit_transform(np.array(comp), W = np.identity(COMPONENTS*INSTRUMENTS), H = np.array(comp))
        print("done")
    model.components_ = np.array(comp)

def transform(filename):
    global model
    if filename.endswith(".wav"):
        fft,phase = get_fft_of_file(filename.replace("\\", "/"))
        prepare_model()
        return model.transform(pack_2d(fft, COMBINED_COLUMNS)),phase

def invtransform(W):
    global model
    prepare_model()
    estimate = model.inverse_transform(W)
    return unpack_2d(estimate)
        
def learn_directory(dir_path, instr_nr, avg_factor = 0.0):
    instrument_complete_dictionary = []
    for file in os.listdir(dir_path):
        filename = os.fsdecode(file)
        if filename.endswith(".wav"): 
            fft_data = prepare_file(os.path.join(dir_path, filename).replace("\\", "/"))

            # Add data to dataset
            for section in fft_data:
                instrument_complete_dictionary.append(section)
    if avg_factor > 0.0:
        instrument_complete_dictionary = limit_array(array_average(instrument_complete_dictionary)*avg_factor)
    print("\n\n\nLearning from "+str(len(instrument_complete_dictionary))+ " segments of data!\n\nLEARNING STARTED! This may take some time...\n")
    model = NMF(n_components=COMPONENTS)
    model.fit_transform(np.array(instrument_complete_dictionary))
    instrumentsH[instr_nr] = normalize_rows(copy.deepcopy(model.components_), 100.0)
    initialized[instr_nr] = True

def array_average(data):
    sum = 0.0
    num = 0.0
    for i in range(len(data)):
        for j in range(len(data[i])):
            sum += data[i][j]
            num += 1.0
    return sum/num

def limit_array(data, limit_value = 1.0):
    for i in range(len(data)):
        for j in range(len(data[i])):
            if data[i][j] < limit_value:
                data[i][j] = 0.0
    return data

def normalize_rows(data, normal_value = 1.0):
    for i in range(len(data)):
        energy_sqrt = 0.0
        for val in data[i]:
            energy_sqrt += val
        if energy_sqrt > 0:
            for j in range(len(data[i])):
                data[i][j] = data[i][j]*normal_value/energy_sqrt
    return data

#
# Combine ajacent columns into packed sections
#
def pack_2d(fft, columns):
    fft_data = []
    # Go through all columns in FFT, except the last few
    for col_i in range(len(fft)+1-columns):
        # Make a new array, copy the curent and several ajacent columns into it
        fft_section = []
        for col_offset in range(columns):
            for i in range(FFT_SIZE):
                fft_section.append(fft[col_i+col_offset][i])
        fft_data.append(fft_section)
    return fft_data

#
# Unpack sections of several columns into one spectrum
#
def unpack_2d(fft_data):
    nr_of_combined_cols = int(len(fft_data[0])/FFT_SIZE)
    fft = []
    # Go through all packed sections
    for i in range(len(fft_data)):
        # unpack, overlap and add sections into single columns
        for col_number in range(nr_of_combined_cols):
            if i+col_number >= len(fft):
                # new column: copy data over to the new column
                new_col = []
                for j in range(FFT_SIZE):
                    new_col.append(fft_data[i][col_number*FFT_SIZE + j])
                fft.append(new_col)
            else:
                # alreadt existing column: add new data to existing data
                for j in range(FFT_SIZE):
                    fft[i+col_number][j] += fft_data[i][col_number*FFT_SIZE + j]
    
    # Divide out number of overlapping columns to get average values
    for i in range(len(fft)):
        for j in range(FFT_SIZE):
            # take care of special cases near the edges of FFT
            if i < nr_of_combined_cols:
                fft[i][j] /= float(i+1);
            elif len(fft)-i < nr_of_combined_cols:
                fft[i][j] /= float(len(fft)-i);
            else:
                fft[i][j] /= float(nr_of_combined_cols);
    return fft


#
# Convert the spectrum data into an array with normalized-energy per window
#
def prepare_file(path):
    print("Preparing file "+path)
    fft,_ = get_fft_of_file(path)
    data = pack_2d(fft, COMBINED_COLUMNS)
    return data

def play_fft(mag, phase):
    global c
    orc = """
    sr=""" + str(SAMPLE_RATE) + """
    ksmps=""" + str(SAMPLES_PER_K) + """
    0dbfs=1

    gifftsize = """ + str(FFT_SIZE) + """
    giFFT ftgen """ + str(TABLE) + """, 0, 0-gifftsize*2, 2, 0

    instr 1
    
    kSpec[] init gifftsize*2
    copyf2array kSpec, giFFT    
    kCmplx[] fftinv kSpec
    kOut[] c2r kCmplx
    
    aout shiftout kOut
    out aout*0.5

    endin
    """
    
    sco = "i1 0 " + str(1+int(len(mag)*FFT_SIZE/SAMPLE_RATE))

    #c.setOption("-odac")
    c.setOption("-o test.wav")
    c.compileOrc(orc)
    c.readScore(sco)
    c.start()
    
    j = 0
    while not c.performKsmps() and j < len(mag):
        fft_complx = []
        for i in range(FFT_SIZE):
            fft_complx.append(mag[j][i]*math.cos(phase[j][i]))
            fft_complx.append(mag[j][i]*math.sin(phase[j][i]))
        
        c.tableCopyIn(TABLE, np.array(fft_complx))
        j = j+1
    c.cleanup()
    c.reset()


def play_fft_mag(fft):
    global c
    orc = """
    sr=""" + str(SAMPLE_RATE) + """
    ksmps=""" + str(SAMPLES_PER_K/4) + """
    0dbfs=1

    gifftsize = """ + str(FFT_SIZE) + """
    giFFT ftgen """ + str(TABLE) + """, 0, 0-gifftsize, 2, 0
    giFFT2 ftgen """ + str(TABLE+1) + """, 0, 0-gifftsize, 2, 0

    instr 1
    ain = 0
    fsigout pvsanal ain, gifftsize, gifftsize/4, gifftsize, 1
    pvsftr fsigout, giFFT, giFFT2
    a2 pvsynth fsigout
    out a2*0.01
    endin
    """
    
    sco = "i1 0 " + str(1+int(len(fft)*FFT_SIZE/SAMPLE_RATE))

    #c.setOption("-odac")
    c.setOption("-o test.wav")
    c.compileOrc(orc)
    c.readScore(sco)
    c.start()
    
    fft_freq = []
    for i in range(FFT_SIZE):
        fft_freq.append(i*SAMPLE_RATE/FFT_SIZE)
    
    j = 0
    while not c.performKsmps() and j < len(fft)*4:
        c.tableCopyIn(TABLE, np.array(fft[int(j/4)]))
        c.tableCopyIn(TABLE+1, np.array(fft_freq))
        j = j+1
    c.cleanup()
    c.reset()    
    
#
# Returns an FFT of an audio-file. Each row is all the bins of a time-window in the FFT
#
# Overlap/window shape is not yet added, but should be easy to add with a few ftab-s.
#
def get_fft_of_file(path):
    global c
    orc = """
    sr=""" + str(SAMPLE_RATE) + """
    ksmps=""" + str(SAMPLES_PER_K) + """
    0dbfs=1

    gifftsize = """ + str(FFT_SIZE) + """
    giFFT ftgen """ + str(TABLE) + """, 0, 0-gifftsize*2, 2, 0

    instr 1
    ilen   filelen \"""" + path + """\"
    xtratim ilen
    
    kIn[] init gifftsize
    a1 diskin2 \"""" + path + """\"

    kCmplx[] r2c kIn
    kSpec[] fft kCmplx
    copya2ftab kSpec, giFFT
    
    kIn[] shiftin a1
    endin
    """
    
    sco = "i1 0 0.1"

    c.setOption("-n")
    c.compileOrc(orc)
    c.readScore(sco)
    c.start()
    
    fft_mag = []
    fft_phase = []
    while not c.performKsmps():
        # Read out complex spectrum, calculate magnitude-spectrum from it
        complex_fft = copy.deepcopy(c.table(TABLE))
        magnitude_fft = []
        phase_fft = []
        for i in range(0, len(complex_fft), 2):
            magnitude_fft.append(math.hypot(complex_fft[i], complex_fft[i+1]))
            phase_fft.append(math.atan2(complex_fft[i+1], complex_fft[i]))
        fft_mag.append(magnitude_fft)
        fft_phase.append(phase_fft)
    c.cleanup()
    c.reset()
    
    return fft_mag,fft_phase

def test_model(instrs, file_list, tresholt):
    global c
    nr = len(instrs)
    
    # Create C-Sound instrument to play samples in different streams
    # Each instrument uses each their stream (global a-rate variable)
    # Streams are FFT'ed individually and mixed, then stored in tables
    # First table is for the mix, the other tables are for individual streams.
    #
    # Samples are played by triggering the instr's.
    #
    orc = """
    sr=""" + str(SAMPLE_RATE) + """
    ksmps=""" + str(SAMPLES_PER_K) + """
    0dbfs=1
    
    gifftsize = """+ str(SAMPLES_PER_K) +"""
    giFFT ftgen """+str(TABLE)+""", 0, 0-gifftsize*2, 2, 0\n"""
    
    for i in range(nr):
        orc += "    gaInst"+str(i+1)+" init 0\n"
        orc += "    giFFT"+str(i+1)+" ftgen "+str(TABLE+i+1)+", 0, 0-gifftsize*2, 2, 0\n"
    orc += "\n"
    for i in range(nr):
        orc += "    instr "+str(i+1)+"\n"
        orc += "        ilen filelen p4\n"
        orc += "        xtratim ilen\n"
        orc += "        a1 diskin2 p4\n"
        orc += "        gaInst"+str(i+1)+" = gaInst"+str(i+1)+" + a1/"+str(nr*2)+".0\n"
        orc += "    endin\n\n"
    
    orc += "    instr 100\n"
    orc += "        aMix = "
    for i in range(nr):
        orc += "gaInst"+str(i+1)+" + "
    orc += """0.0
        
        kIn[] init gifftsize
        kCmplx[] r2c kIn
        kSpec[] fft kCmplx
        copya2ftab kSpec, giFFT\n\n"""
    
    for i in range(nr):    
        orc += "        kIn"+str(i+1)+"[] init gifftsize\n"
        orc += "        kCmplx"+str(i+1)+"[] r2c kIn"+str(i+1)+"\n"
        orc += "        kSpec"+str(i+1)+"[] fft kCmplx"+str(i+1)+"\n"
        orc += "        copya2ftab kSpec"+str(i+1)+", giFFT"+str(i+1)+"\n\n"

    for i in range(nr):
        orc += "        kIn"+str(i+1)+"[] shiftin gaInst"+str(i+1)+"\n"
        orc += "        gaInst"+str(i+1)+" = 0\n\n"
    orc += "        kIn[] shiftin aMix\n"
    orc += "    endin\n"
    #############################################################
    
    sco = "i100 0 "+str(TEST_TOTAL_TIME+5.0)+"\n"
    for i in range(int(TEST_TOTAL_TIME*TEST_SAMPLE_DENSITY)):
        for j in range(nr):
            sco += "i"+str(j+1)+" "+str(randrange(int(TEST_TOTAL_TIME*100))/100.0)+" 0.1 \""+file_list[instrs[j]][randrange(len(file_list[instrs[j]]))]+"\"\n"
        
    #############################################################

    c.setOption("-n")
    c.compileOrc(orc)
    c.readScore(sco)
    c.start()

    # Read out complex spectrums, calculate magnitude-spectrums from it
    fft_mix = []
    fft_streams = [[] for i in range(nr)]
    while not c.performKsmps():
    
        # Mixed
        complex_fft = copy.deepcopy(c.table(TABLE))
        magnitude_fft = []
        for i in range(0, len(complex_fft), 2):
            magnitude_fft.append(math.hypot(complex_fft[i], complex_fft[i+1]))
        fft_mix.append(magnitude_fft)
        
        # Separate
        for i in range(nr):
            complex_fft = copy.deepcopy(c.table(TABLE+1+i))
            magnitude_fft = []
            for j in range(0, len(complex_fft), 2):
                magnitude_fft.append(math.hypot(complex_fft[j], complex_fft[j+1]))
            fft_streams[i].append(magnitude_fft)
        
    c.cleanup()
    c.reset()
       
    prepare_model()
    
    print("\n\nAnalyzing...\n")
    W = model.transform(pack_2d(fft_mix, COMBINED_COLUMNS))

    print("\tWhole mix")
    fft_derived_mix = invtransform(copy.deepcopy(W))

    #
    # Use model to separate streams from mix
    #
    fft_derived_streams_pruning = [[] for i in range(nr)] 
    fft_derived_streams_no_pruning = [[] for i in range(nr)]
    fft_derived_streams_reclass = [[] for i in range(nr)]
    for inst_no in range(nr):
        print("\tStream",inst_no+1,"out of", nr)
        W_prune = copy.deepcopy(W)
        W_no_prune = copy.deepcopy(W)
        W_reclass = copy.deepcopy(W)
        
        for i in range(INSTRUMENTS):
            for j in range(COMPONENTS):
                if componentMetrics[i*COMPONENTS + j][i] < tresholt:
                    for k in range(len(W)):
                        W_reclass[k][i*COMPONENTS + j] = 0.0
                if i != instrs[inst_no]:
                    for k in range(len(W)):
                        W_prune[k][i*COMPONENTS + j] = 0.0
                        W_no_prune[k][i*COMPONENTS + j] = 0.0
                elif componentMetrics[i*COMPONENTS + j][i] < tresholt:
                    for k in range(len(W)):
                        W_prune[k][i*COMPONENTS + j] = 0.0
        
        fft_derived_streams_no_pruning[inst_no] = invtransform(W_no_prune)
        fft_derived_streams_pruning[inst_no] = invtransform(W_prune)
        fft_derived_streams_reclass[inst_no] = invtransform(W_reclass)

    
    print("")
    results = []
    for inst_no in range(nr):
        print("calculating statistics for stream", inst_no+1)
        diff_original_mix = 0.0
        diff_derived_mix = 0.0
        diff_reclassified = 0.0
        diff_no_pruning = 0.0
        diff_pruning = 0.0
        total_level = 0.0
        for i in range(len(fft_streams[inst_no])):
            for j in range(len(fft_streams[inst_no][i])):
                diff_original_mix += abs(fft_streams[inst_no][i][j] - fft_mix[i][j])
                diff_derived_mix += abs(fft_streams[inst_no][i][j] - fft_derived_mix[i][j])
                diff_no_pruning += abs(fft_streams[inst_no][i][j] - fft_derived_streams_no_pruning[inst_no][i][j])
                diff_pruning += abs(fft_streams[inst_no][i][j] - fft_derived_streams_pruning[inst_no][i][j])
                diff_reclassified += abs(fft_streams[inst_no][i][j] - fft_derived_streams_reclass[inst_no][i][j])
                total_level += fft_streams[inst_no][i][j]
                
        results.append((instrs[inst_no], total_level, diff_original_mix, diff_derived_mix, diff_no_pruning, diff_pruning, diff_reclassified, copy.deepcopy(instrs)))
    return results

def add_classification_data(test_data, componentMetrics):
    prepare_model()
    for i in range(INSTRUMENTS):
        if len(test_data[i]) > 0:
            print("Transforming samples of instrument nr", i)
            W = model.transform(np.array(test_data[i]))
            print("Adding data from",len(W),"segments")    
            # Add together the components contribution for classifying a sample of a known instrument
            for j in range(COMPONENTS*INSTRUMENTS):
                for k in range(len(W)):
                    componentMetrics[j][i] += W[k][j]
        else:
            print("No data from instrument", i)
    return componentMetrics

def fresh_input(text):
    while kbhit(): getch()
    return input(text)

root = tk.Tk()
root.withdraw()

while 1==1:
    print("""\nActions:
    0. Clear everything and change parameters
    1. Transform/inverse-transform Audio using model
    2. Source separation using classification only
    3. Calculate Instrument Utilization Ratios (make pruning data)
    4. Source separation using classification and pruning
    5. Verification test of final model
    6. Save learned instrument
    7. Restore saved instrument
    8. Restore instrument utilization data
    9. Learn new insrument from folder\n""")
    
    choice = int(fresh_input("Enter choice: "))
    if choice == 0:
        COMPONENTS = int(fresh_input("Features per instrument in model (curently "+str(COMPONENTS)+"): "))
        #FFT_SIZE = int(fresh_input("FFT-size (curently "+str(FFT_SIZE)+")? "))
        #COMBINED_COLUMNS = int(fresh_input("Feature length (in FFT-windows, curently "+str(COMBINED_COLUMNS)+")? "))
        
        instrumentsH = [[([1.0]*FFT_SIZE*COMBINED_COLUMNS) for i in range(COMPONENTS)] for j in range(INSTRUMENTS)]
        initialized = [False]*INSTRUMENTS
        componentMetrics = [([100.0]*INSTRUMENTS) for i in range(COMPONENTS*INSTRUMENTS)]
        model = None
    elif choice == 1:
        print("\n\nPlease select an audio file to transform...")
        file = filedialog.askopenfile()
        W,phase = transform(file.name)
        
        # Calculate percentage of each instrument usage
        instr_feat_amp = [0.0]*INSTRUMENTS
        total_feat_amp = 0.0
        for i in range(INSTRUMENTS):
            for j in range(COMPONENTS):
                for k in range(len(W)):
                    instr_feat_amp[i] += W[k][i*COMPONENTS + j]
                    total_feat_amp += W[k][i*COMPONENTS + j]
        for i in range(INSTRUMENTS):
            instr_feat_amp[i] = instr_feat_amp[i]*100.0/total_feat_amp
        print("Percentage use per instrument:", instr_feat_amp, "\n")
                    
        play_fft(invtransform(W), phase)
        print("\n\nSaved to test.wav\n")
    elif choice == 2:
        
        file = filedialog.askopenfile()
        instr_nr = int(fresh_input("Instrument number to separate (0-"+str(INSTRUMENTS-1)+"): "))
        if instr_nr >= 0 and instr_nr < INSTRUMENTS:
            W,phase = transform(file.name)
            
            # remove all other instruments
            for i in range(INSTRUMENTS):
                if i != instr_nr:
                    for j in range(COMPONENTS):
                        for k in range(len(W)):
                            W[k][i*COMPONENTS + j] = 0.0
        
            play_fft(invtransform(W), phase)
            #play_fft_mag(invtransform(W))
            print("\n\nSaved to test.wav\n")
    elif choice == 3:
        componentMetrics = [([0.0]*INSTRUMENTS) for i in range(COMPONENTS*INSTRUMENTS)]
        test_data = [[] for i in range(INSTRUMENTS)]
        accumulated_files = 0
        print("\n\nPlease select a directory containing the entire training-set...")
        dir = filedialog.askdirectory()
        for file in os.listdir(dir):
            filename = os.fsdecode(file)
            if filename.endswith(".wav"):
                instr_nr = int(filename[:1])    # This is where classification data is gathered for the audio in the training set during this operation
                
                if initialized[instr_nr]:
                    # Add data to dataset
                    fft_data = prepare_file(os.path.join(dir, filename).replace("\\", "/"))
                    accumulated_files += 1
                    for section in fft_data:
                        test_data[instr_nr].append(section)
                        
                    if accumulated_files > 100:
                        componentMetrics = add_classification_data(test_data, componentMetrics)
                        test_data = [[] for i in range(INSTRUMENTS)]
                        accumulated_files = 0
        
        if accumulated_files > 0:
            componentMetrics = add_classification_data(test_data, componentMetrics)
        
        # Normalize contributions from individual instruments
        for i in range(INSTRUMENTS):
            sum = 0.0
            for j in range(COMPONENTS*INSTRUMENTS):
                sum += componentMetrics[j][i]
            if sum > 0.0:
                for j in range(COMPONENTS*INSTRUMENTS):
                    componentMetrics[j][i] = componentMetrics[j][i]*COMPONENTS*INSTRUMENTS*100.0/sum
        
        # Convert contributions to percentage for each feature
        for i in range(COMPONENTS*INSTRUMENTS):
            sum = 0.0
            for j in range(INSTRUMENTS):
                sum += componentMetrics[i][j]
            if sum > 0.0:
                for j in range(INSTRUMENTS):
                    componentMetrics[i][j] = componentMetrics[i][j]*100.0/sum
        
        print("")
        for b in initialized:
            print(b)
        print("")
        
        print("\n\nPlease chose file to save Instrument Utilization Data to...")
        # save statistics to file
        file = filedialog.asksaveasfilename()
        with open(file, "w") as wf:
            for row in componentMetrics:
                for val in row:
                    wf.write(str(val))
                    wf.write(" ")
                wf.write("\n")
            wf.close()
    elif choice == 4:
        print("\n\nPlease select an audio file to source-separate...")
        file = filedialog.askopenfile()
        instr_nr = int(fresh_input("Instrument number (0-"+str(INSTRUMENTS-1)+"): "))
        if instr_nr >= 0 and instr_nr < INSTRUMENTS:
            # ask for tresholt %
            tresholt = float(fresh_input("Prune components with a classification accuracy below what percentage (0 - 100)? "))
            
            W,phase = transform(file.name)
            
            # remove all other instruments and prune
            for i in range(INSTRUMENTS):
                if i != instr_nr:
                    for j in range(COMPONENTS):
                        for k in range(len(W)):
                            W[k][i*COMPONENTS + j] = 0.0
                else:
                    nr = 0
                    for j in range(COMPONENTS):
                        if componentMetrics[i*COMPONENTS + j][i] < tresholt:
                            for k in range(len(W)):
                                W[k][i*COMPONENTS + j] = 0.0
                            nr += 1
                    print("Pruned",nr,"patterns in instrument",i)        
        
            play_fft(invtransform(W), phase)
            #play_fft_mag(invtransform(W))
            print("\n\nSaved to test.wav\n")
    elif choice == 5:
        
        nr = int(input("Number of instruments per combination (2-"+str(initialized.count(True))+")? "))
        nr_combos = int(input("Number of repeated tests per instrument combination? "))
        
        
        if nr > 1 and nr_combos >0 and initialized.count(True) >= nr:
            # ask for tresholt %
            tresholt = float(fresh_input("Prune components with a positive hitrate below what percentage? "))

            results = []
            
            print("\n\nPlease select a directory containing the entire test-set...")
            # Get list of files
            dir = filedialog.askdirectory()
            file_list = [[] for i in range(INSTRUMENTS)]
            for file in os.listdir(dir):
                filename = os.fsdecode(file)
                if filename.endswith(".wav"):
                    instr_nr = int(filename[:1])    # This is where classification data is gathered for the audio in the test set during this operation
                    file_list[instr_nr].append(os.path.join(dir, filename).replace("\\", "/"))
            
            for combination in combinations(range(INSTRUMENTS), nr): 
                dotest = True
                for i in combination:
                    if not initialized[i]:
                        dotest = False
                
                if dotest:
                    local_results = []
                    for i in range(nr_combos):
                        local_results.extend(test_model(combination, file_list, tresholt))
            
                    results.append(local_results)
            
            print("\nHas been pruning below ",tresholt,"%\n\nAll data is in the order mix, transform, class separation, class+pruned separation and pruned separation\n")
            
            #
            # Format of results is lists of results from every combination. Each combination-results has a list of the following tuple:
            #
            #   * instrument number for the stream
            #   * total magnitude-sum of original instrument stream
            #   * magnitude difference with the non-altered mix
            #   * magnitude difference with the transformed whole mix
            #   * magnitude difference with separation, non-pruning with pre-classification
            #   * magnitude difference with separation, pruning with pre-classification
            #   * magnitude difference with separation, pruning without pre-classification
            #   * list of other instruments used in test
            #
            
            overall_sum_mix = [0.0] * INSTRUMENTS
            overall_sum_mix_t = [0.0] * INSTRUMENTS
            overall_sum_nonpruning = [0.0] * INSTRUMENTS
            overall_sum_pruning = [0.0] * INSTRUMENTS
            overall_sum_reclass = [0.0] * INSTRUMENTS
            overall_num = [0] * INSTRUMENTS
            overall_avg_mix = [0.0] * INSTRUMENTS
            overall_avg_mix_t = [0.0] * INSTRUMENTS
            overall_avg_nonpruning = [0.0] * INSTRUMENTS
            overall_avg_pruning = [0.0] * INSTRUMENTS
            overall_avg_reclass = [0.0] * INSTRUMENTS
            overall_var_mix = [0.0] * INSTRUMENTS
            overall_var_mix_t = [0.0] * INSTRUMENTS
            overall_var_nonpruning = [0.0] * INSTRUMENTS
            overall_var_pruning = [0.0] * INSTRUMENTS
            overall_var_reclass = [0.0] * INSTRUMENTS
            
            for single_comb_results in results:
                combination_sum_mix = [0.0] * INSTRUMENTS
                combination_sum_mix_t = [0.0] * INSTRUMENTS
                combination_sum_nonpruning = [0.0] * INSTRUMENTS
                combination_sum_pruning = [0.0] * INSTRUMENTS
                combination_sum_reclass = [0.0] * INSTRUMENTS
                combination_num = [0] * INSTRUMENTS
                combination_avg_mix = [0.0] * INSTRUMENTS
                combination_avg_mix_t = [0.0] * INSTRUMENTS
                combination_avg_nonpruning = [0.0] * INSTRUMENTS
                combination_avg_pruning = [0.0] * INSTRUMENTS
                combination_avg_reclass = [0.0] * INSTRUMENTS
                combination_var_mix = [0.0] * INSTRUMENTS
                combination_var_mix_t = [0.0] * INSTRUMENTS
                combination_var_nonpruning = [0.0] * INSTRUMENTS
                combination_var_pruning = [0.0] * INSTRUMENTS
                combination_var_reclass = [0.0] * INSTRUMENTS

                
                # Local Mean
                for datapoint in single_comb_results:
                    combination_sum_mix[datapoint[0]] += datapoint[2]/datapoint[1]
                    combination_sum_mix_t[datapoint[0]] += datapoint[3]/datapoint[1]
                    combination_sum_nonpruning[datapoint[0]] += datapoint[4]/datapoint[1]
                    combination_sum_pruning[datapoint[0]] += datapoint[5]/datapoint[1]
                    combination_sum_reclass[datapoint[0]] += datapoint[6]/datapoint[1]
                    combination_num[datapoint[0]] += 1

                
                for i in range(INSTRUMENTS):
                    if combination_num[i] > 0:
                        combination_avg_mix[i] = combination_sum_mix[i]/combination_num[i]
                        combination_avg_mix_t[i] = combination_sum_mix_t[i]/combination_num[i]
                        combination_avg_nonpruning[i] = combination_sum_nonpruning[i]/combination_num[i]
                        combination_avg_pruning[i] = combination_sum_pruning[i]/combination_num[i]
                        combination_avg_reclass[i] = combination_sum_reclass[i]/combination_num[i]
                        print("Avg error of ", i, " from", single_comb_results[0][7], "is", combination_avg_mix[i], combination_avg_mix_t[i], combination_avg_nonpruning[i], combination_avg_pruning[i], combination_avg_reclass[i])
                        
                        overall_sum_mix[i] += combination_sum_mix[i]
                        overall_sum_mix_t[i] += combination_sum_mix_t[i]
                        overall_sum_nonpruning[i] += combination_sum_nonpruning[i]
                        overall_sum_pruning[i] += combination_sum_pruning[i]
                        overall_sum_reclass[i] += combination_sum_reclass[i]
                        overall_num[i] += combination_num[i]
                
                # Local Standard Deviation
                for datapoint in single_comb_results:
                    combination_var_mix[datapoint[0]] += ((datapoint[2]/datapoint[1])-combination_avg_mix[datapoint[0]])**2
                    combination_var_mix_t[datapoint[0]] += ((datapoint[3]/datapoint[1])-combination_avg_mix_t[datapoint[0]])**2
                    combination_var_nonpruning[datapoint[0]] += ((datapoint[4]/datapoint[1])-combination_avg_nonpruning[datapoint[0]])**2
                    combination_var_pruning[datapoint[0]] += ((datapoint[5]/datapoint[1])-combination_avg_pruning[datapoint[0]])**2
                    combination_var_reclass[datapoint[0]] += ((datapoint[6]/datapoint[1])-combination_avg_reclass[datapoint[0]])**2
                    
                for i in range(INSTRUMENTS):
                    if combination_num[i] > 0:
                        print("Error's SD of", i, "from", single_comb_results[0][7], "is", combination_var_mix[i]/(combination_num[i]-1.0), combination_var_mix_t[i]/(combination_num[i]-1.0), combination_var_nonpruning[i]/(combination_num[i]-1.0), combination_var_pruning[i]/(combination_num[i]-1.0), combination_var_reclass[i]/(combination_num[i]-1.0))
            
            # Total Mean
            for i in range(INSTRUMENTS):
                if initialized[i]:
                    overall_avg_mix[i] = overall_sum_mix[i]/overall_num[i]
                    overall_avg_mix_t[i] = overall_sum_mix_t[i]/overall_num[i]
                    overall_avg_nonpruning[i] = overall_sum_nonpruning[i]/overall_num[i]
                    overall_avg_pruning[i] = overall_sum_pruning[i]/overall_num[i]
                    overall_avg_reclass[i] = overall_sum_reclass[i]/overall_num[i]
                    print("Average error of", i, "is", overall_avg_mix[i], overall_avg_mix_t[i], overall_avg_nonpruning[i], overall_avg_pruning[i], overall_avg_reclass[i])

            # Total Standard Deviation
            for single_comb_results in results:
                for datapoint in single_comb_results:
                    overall_var_mix[datapoint[0]] += ((datapoint[2]/datapoint[1])-overall_avg_mix[datapoint[0]])**2
                    overall_var_mix_t[datapoint[0]] += ((datapoint[3]/datapoint[1])-overall_avg_mix_t[datapoint[0]])**2
                    overall_var_nonpruning[datapoint[0]] += ((datapoint[4]/datapoint[1])-overall_avg_nonpruning[datapoint[0]])**2
                    overall_var_pruning[datapoint[0]] += ((datapoint[5]/datapoint[1])-overall_avg_pruning[datapoint[0]])**2
                    overall_var_reclass[datapoint[0]] += ((datapoint[6]/datapoint[1])-overall_avg_reclass[datapoint[0]])**2
                    
            for i in range(INSTRUMENTS):
                if initialized[i]:
                    print("Error's SD of", i, "is", overall_var_mix[i]/(overall_num[i]-1.0), overall_var_mix_t[i]/(overall_num[i]-1.0), overall_var_nonpruning[i]/(overall_num[i]-1.0), overall_var_pruning[i]/(overall_num[i]-1.0), overall_var_reclass[i]/(overall_num[i]-1.0))
            
    elif choice == 6:
        instr_nr = int(fresh_input("Instrument number (0-"+str(INSTRUMENTS-1)+"): "))
        if instr_nr >= 0 and instr_nr < INSTRUMENTS and initialized[instr_nr] == True:
            file = filedialog.asksaveasfilename()
            print("Saving instrument", instr_nr, "as file", file)
            with open(file, "w") as wf:
                for component in instrumentsH[instr_nr]:
                    for val in component:
                        wf.write(str(val))
                        wf.write(" ")
                    wf.write("\n")
                wf.close()
    elif choice == 7:
        instr_nr = int(fresh_input("Instrument number (0-"+str(INSTRUMENTS-1)+"): "))
        if instr_nr >= 0 and instr_nr < INSTRUMENTS:
            file = filedialog.askopenfilename()
            print("Loading instrument", instr_nr, "from file", file)
            loaded_instrument = []
            with open(file, "r") as rf:
                for line in rf:
                    line = line.strip()
                    loaded_instrument.append([float(x) for x in line.split()])
                rf.close()
            instrumentsH[instr_nr] = loaded_instrument
            initialized[instr_nr] = True
    elif choice == 8:
        file = filedialog.askopenfilename()
        print("Loading file", file, "as pruning data")
        loaded_data = []
        with open(file, "r") as rf:
            for line in rf:
                line = line.strip()
                loaded_data.append([float(x) for x in line.split()])
            rf.close()
        componentMetrics = loaded_data
    elif choice == 9:
        instr_nr = int(fresh_input("Instrument number (0-"+str(INSTRUMENTS-1)+"): "))
        if instr_nr >= 0 and instr_nr < INSTRUMENTS:
            #percent_limit = float(fresh_input("Limit below (average * x%): "))
            percent_limit = 0.0
            print("\n\nPlease select a directory containing ONLY the training-set for one single instrument class...")
            dir = filedialog.askdirectory(initialdir="C:\\Users\\Frode\\Desktop\\")
            learn_directory(dir, instr_nr, percent_limit/100.0)
            print("Done learning instrument",instr_nr)

#plt.imshow(phrase)
#plt.show()